#ifndef CONTROL_EVENT_DOT_H
#define CONTROL_EVENT_DOT_H

/* ====================================================================
 * The Vovida Software License, Version 1.0
 *
 * Copyright (c) 1999, 2000 Vovida Networks, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by Vovida Networks,
 *        Inc. (http://www.vovida.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must 
 *    not be used to endorse or promote products derived from this 
 *    software without prior written permission. For written permission, 
 *    please contact vocal@vovida.org.
 *
 * 5. Products derived from this software may not be called "VOCAL",
 *    nor may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL VOVIDA NETWORKS, INC. OR ITS 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by Vovida 
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see 
 * <http://www.vovida.org/>.
 *
 */





static const char* const ControlEventHeaderVersion = 
    "$Id: ControlEvent.hxx,v 1.5 2000/08/10 11:16:54 bogawa Exp $";




#include "Event.hxx"
#include "Sptr.hxx"
#include "FifoBase.h"
#include "VMutex.h"

namespace Vovida
{

/// ControlEvent is used to manage the starting, stopping and shutting down of
/// thread that wait on Fifo<Event>'s. It is best to only use these within
/// an Sptr.
///
class ControlEvent : public Event
{
    public:

    	/// The type of control event.
	///
    	enum Type
	{
	    START = 0,
	    STOP,
	    SHUTDOWN
	};

    	/// Create a control event with a type and an optional reply
	/// fifo. If the reply fifo is set, the control event can be
	/// used as a request (rather than a command), and the response
	/// can be communicated via the returncode.
	///
    	ControlEvent(Type, FifoBase< Sptr<Event> > * = 0);
    	ControlEvent(Type, FifoBase< Sptr<Event> > &);

	virtual     ~ControlEvent();


    	virtual const char* const   	name() const;


    	Type	    	    	    	getType() const;
		

    	void	    	    	    	setReturnCode(int);
    	int 	    	    	    	getReturnCode() const;
		
    	void	    	    	    	setReplyFifo(
	    	    	    	    	    FifoBase< Sptr<Event> > &
					);
    	FifoBase< Sptr<Event> > *   	getReplyFifo() const;

    	/// Sends the event back to the reply fifo, if set.
	///
    	/// WARNING: This assumes that this is already housed in an Sptr.
	///  	     Don't use this method if it is on the stack, or it can be
	/// 	     manually deleted, since it will cause a double delete.
	/// 	
    	virtual void	    	    	returnReply();
			
    	/// To uniquely identify a request. Useful if this control event is
	/// a request. The fifo waiting for the response can the uniquely
	/// identify the response, in the case they sent out multiple control
	/// events simultaneously.
	///
    	int 	    	    	    	getControlEventId() const;

    	/// Write the event to a stream.
	///
	virtual ostream &   	    	writeTo(ostream &) const;

    private:
    	/// nextId is a globally shared resource, so it needs to be protected.
	///
    	static	VMutex	    	    	nextIdMutex;
    	static	int 	    	    	nextId;

	Type	    	    	    	type;
        int 	    	    	    	returnCode;
	FifoBase< Sptr<Event> >     *	fifo;
    	int 	    	    	    	id;

    	/// Suppress copying.
	///
    	ControlEvent(const ControlEvent &);
	ControlEvent & operator=(const ControlEvent &);
};

} // namespace Vovida

using namespace Vovida;

 // !defined(CONTROL_EVENT_DOT_H)
#endif
