#ifndef FIFOBASE_DOT_H
#define FIFOBASE_DOT_H

/* ====================================================================
 * The Vovida Software License, Version 1.0
 *
 * Copyright (c) 1999, 2000 Vovida Networks, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by Vovida Networks,
 *        Inc. (http://www.vovida.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must 
 *    not be used to endorse or promote products derived from this 
 *    software without prior written permission. For written permission, 
 *    please contact vocal@vovida.org.
 *
 * 5. Products derived from this software may not be called "VOCAL",
 *    nor may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL VOVIDA NETWORKS, INC. OR ITS 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by Vovida 
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see 
 * <http://www.vovida.org/>.
 *
 */





static const char* const FifoBaseHeaderVersion = 
    "$Id: FifoBase.h,v 1.9 2000/08/10 11:16:55 bogawa Exp $";




#include "VMutex.h"
#include "VCondition.h"
#include "TimerContainer.hxx"
#include "Sptr.hxx"
#include <list>

template<class Msg>
class FifoBase 
{
    public:

    	/// Id for delayed events. Needed to cancel an event. 
	///
    	typedef TimerEntryId 	    	EventId;

    protected:	
    
	/// Create an empty fifo.
	///
    	FifoBase();

    public:
    
	/// Delete the fifo.
	///
    	virtual     	    	    ~FifoBase();

	/// Add a message to the fifo.
	///
    	void	     	    	    add(const Msg &);

    	/// For backward compatibility. If the relative timeout
	/// (in milliseconds) is greater than 0, the message will
	/// be added to the fifo after the number of milliseconds
	/// have passed.
	///
	/// The returned value is an opaque id that can be used
	/// to cancel the event before the timer expires. If the 
	/// time is specified as 0, this defaults the to the 
	/// regular add(), returning an id of 0.
	///
    	EventId     	    	    addDelayMs(
	    	    	    	    	const Msg &, 
	    	    	    	    	const milliseconds_t relativeTimeout);

	/// Cancel a delayed event. 
	///
	void 	    	    	    cancel(EventId);

    	/// Blocks until a message is available, or the specified timeout has
	/// expired. If the return code is 0, then the fifo was interrupted 
	/// via a signal or the timeout expired without a message being added
	/// to the queue. Using block is the only way to guarantee detection 
	/// of a signal has been thrown. 
	///
    	int 	    	    	    block(milliseconds_t relativeTimeout = -1) 
	    	    	    	    	throw ( VException );
	
	/// Returns the first message available. It will wait if no 
	/// messages are available. If a signal interrupts the wait,
	/// it will retry the wait. Signals can therefore not be caught
	/// via getNext. If you need to detect a signal, use block
	/// prior to calling getNext.
	///
	Msg 	    	    	    getNext() throw ( VException );

        /** Get the current size of the FIFO Note that the current
	    size INCLUDES all of the pending events, even those which
	    have not yet activated so you should NOT use this function
	    to determine whether a call to getNext will block or not.
	    Use messageAvailable() instead.
	*/
        unsigned int                size() const;

    	/// Returns true if a message is available. 
	///
    	bool	    	    	    messageAvailable();

    	/// Indicate that the poll no longer accepts messages.
	///
    	void	    	    	    shutdown();

    	/// Relational operators. Useful if this is ever in a container.
	///		
	bool	    	    	    operator==(const FifoBase &) const;
	bool	    	    	    operator!=(const FifoBase &) const;
	bool	    	    	    operator< (const FifoBase &) const;
	bool	    	    	    operator<=(const FifoBase &) const;
	bool	    	    	    operator> (const FifoBase &) const;
	bool	    	    	    operator>=(const FifoBase &) const;
    	    			
    protected:

    	/// Returns true if a message is available. The default implementation
	/// looks at the size of the fifo. Derived classes may also observe
	/// other resources as well. Note that the mutex should not be locked
	/// by this routine, since it would cause a deadlock.
	///
    	virtual bool	    	    messageAvailableNoLock();
    
    	/// Wake's up the blocked thread. Note that the mutex is locked
	/// upon entry. It is the responsibility of the user to unlock 
	/// the mutex if necessary, making sure to relock it before exit.
	///
    	virtual void	    	    wakeup() throw ( VException ) = 0;

    	/// Blocks the thread, for the given time, waiting for a message. 
	/// Note that the mutex is locked upon entry. It is the responsibility
	/// of the user to unlock the mutex if necessary, making sure
	/// to relock it before exit. The return code is positive for 
	/// to indicate activity, i.e. messages added to the queue. For
	/// a signal or a timeout, 0 should be returned.
	///
	virtual int	    	    wait(milliseconds_t relativeTimeout) 
	    	    	    	    	throw ( VException ) = 0;


    	// Protects fifo and timer.
	//
	typedef     list< Sptr<Msg> >	MessageContainer;
	
	MessageContainer  	    	fifo_;
	mutable VMutex	    	    	mutex_; 	
	
    private:

    	int 	    	    	    blockNoLock(milliseconds_t relativeTimeout = -1) 
	    	    	    	    	throw ( VException );
	
	Vovida::TimerContainer<Msg> 	timer_;
	bool	    	    	    	shutdown_;

    	/// Suppress copying
	///
	FifoBase(const FifoBase &);
	FifoBase & operator=(const FifoBase &);
};

// For backwards compatibility.
//
typedef TimerEntryId 	    	    FifoEventId;

#include "FifoBase.cc"

// !defined(FIFOBASE_DOT_H)
#endif
