
/* ====================================================================
 * The Vovida Software License, Version 1.0
 *
 * Copyright (c) 1999, 2000 Vovida Networks, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by Vovida Networks,
 *        Inc. (http://www.vovida.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must 
 *    not be used to endorse or promote products derived from this 
 *    software without prior written permission. For written permission, 
 *    please contact vocal@vovida.org.
 *
 * 5. Products derived from this software may not be called "VOCAL",
 *    nor may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL VOVIDA NETWORKS, INC. OR ITS 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by Vovida 
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see 
 * <http://www.vovida.org/>.
 *
 */

static const char* const TimerEntry_cc_Version = 
"$Id: TimerEntry.cc,v 1.7 2000/08/10 11:16:56 bogawa Exp $";


#include <cassert>

using namespace Vovida;


template<class Msg>
TimerEntry<Msg>::TimerEntry(Sptr<Msg> msg, milliseconds_t relativeTime)
    :	msg_(msg)
{
    int rc = gettimeofday(&absExpires_, 0);
    assert( rc == 0 );

    if ( relativeTime < 0 )
    {
    	relativeTime = 0;
    }

    absExpires_.tv_sec	+= relativeTime / 1000;
    absExpires_.tv_usec += (relativeTime % 1000) * 1000;
}



template<class Msg>
TimerEntry<Msg>::TimerEntry(const TimerEntry & src) 
    :	absExpires_(src.absExpires_),
	msg_(src.msg_)
{
}


template<class Msg>
TimerEntry<Msg>::~TimerEntry()
{
}
    	

template<class Msg>
TimerEntry<Msg>::TimerEntry &	    
TimerEntry<Msg>::operator=(const TimerEntry & src)
{
    if ( this != &src )
    {
	absExpires_ = src.absExpires_;
	msg_ = src.msg_;
    }
    return ( *this );
}


template<class Msg>
milliseconds_t	    
TimerEntry<Msg>::getTimeout() const
{
    timeval now;
    int rc = gettimeofday(&now, 0);
    assert( rc == 0 );
    
    milliseconds_t timeout  = ( absExpires_.tv_sec - now.tv_sec ) * 1000
    	    	    	    + ( absExpires_.tv_usec - now.tv_usec ) / 1000;

    // If now is after the absolute expiry time, the timeout is negative and
    // we should return a timeout value of 0.
    //			    
    return ( timeout < 0 ? 0 : timeout );
}


template<class Msg>
TimerEntryId
TimerEntry<Msg>::getId() const
{
    return ( (TimerEntryId)(msg_.operator->()) );
}


template<class Msg>
Sptr<Msg>
TimerEntry<Msg>::getMessage() const
{
    return ( msg_ );
}


template<class Msg>
bool	    	    
TimerEntry<Msg>::hasExpired() const
{
    timeval now;
    int rc = gettimeofday(&now, 0);
    assert( rc == 0 );

    bool expired = false;
    
    // We have expired if the absExpiry time is less than or 
    // equal to the time now.
    //

    if ( absExpires_.tv_sec < now.tv_sec )
    {
    	expired = true;
    }
    else if ( 	absExpires_.tv_sec == now.tv_sec   
    	    &&  absExpires_.tv_usec <= now.tv_usec 
	    )
    {
    	expired = true;
    }    
	    
    return ( expired );
}


template<class Msg>
bool	    	    
TimerEntry<Msg>::operator==(const TimerEntry & src) const
{
    return  ( 	absExpires_.tv_sec == src.absExpires_.tv_sec 
    	    &&  absExpires_.tv_usec == src.absExpires_.tv_usec 
	    );
}


template<class Msg>
bool	    	    
TimerEntry<Msg>::operator!=(const TimerEntry & src) const
{
    return ( !operator==(src) );
}


template<class Msg>
bool	    	    
TimerEntry<Msg>::operator<(const TimerEntry & src) const
{
    if ( absExpires_.tv_sec < src.absExpires_.tv_sec )
    {
        return ( true );
    } 
    else if ( absExpires_.tv_sec > src.absExpires_.tv_sec )
    { 
        return ( false );
    }
    else
    {
        return ( absExpires_.tv_usec < src.absExpires_.tv_usec );
    }
}


template<class Msg>
bool	    	    
TimerEntry<Msg>::operator<=(const TimerEntry & src) const
{
    return ( operator<(src) || operator==(src) );
}


template<class Msg>
bool	    	    
TimerEntry<Msg>::operator>(const TimerEntry & src) const
{
    return ( !operator<=(src) );
}


template<class Msg>
bool	    	    
TimerEntry<Msg>::operator>=(const TimerEntry & src) const
{
    return ( !operator<(src) );
}


template<class Msg>
ostream &   
TimerEntry<Msg>::writeTo(ostream & out) const
{
    return  ( out   << "id = " << getId()
    	    	    << ", absolute time of expiration = "
		    << "{ sec: "  << absExpires_.tv_sec 
		    << ", usec: " << absExpires_.tv_usec << " }"
		    << ", relative time of expiration = " << getTimeout()
	    );
}
